//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

/** CPCTelera includes */
#include <cpctelera.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Game includes */
#include "..\view\view.h"
#include "..\tools\tools.h"

/** Level */
#include ".\level.h"

/** Local structures */
typedef void(*TCreateLevel)();

/** Local variables */
static u8 _populate;

/** Extern variables */
extern SDungeon gDungeon;

/** Local defines */
#define HORIZ	TRUE
#define VERT	FALSE

/*****************************************************/
/*													 */
/*	Get random range								 */
/*												     */
/*****************************************************/
u8 GetRandRange(u8 from, u8 size)
{
	return from + GetRandModulo(size);
}

/*****************************************************/
/*													 */
/*	Add level items		 							 */
/*												     */
/*****************************************************/
void AddLevelItems(u8 nb, u8 item, u8 range)
{
	for (u8 y = 2; y < MAP_CY - 1; y++)
	{
		for (u8 x = 1; x < MAP_CX - 1; x++)
		{
			u8* tileMap = GetLevelPtr(x, y);
		
			if (*tileMap == DGN_GROUND && GetRandModulo(MAP_SIZE / nb) == 0)
				*tileMap = item + GetRandModulo(range);
		}
	}
}

/*****************************************************/
/*													 */
/*	Insert doors in Wall							 */
/*												     */
/*****************************************************/
void InsertDoorInWall(u8 x, u8 y, u8 doorId)
{
	u8* map = GetLevelPtr(x, y);
	u8 offset = MAP_CX;	
	u8 wallBorder = 4;
	
	if (doorId == DGN_DOOR_H_RED || doorId == DGN_DOOR_H_BLUE)
	{
		offset = 1;
		wallBorder = 3;
	}
	
	*map = WALLS_BEG + wallBorder;
	*(map + offset) = doorId;
	*(map + offset*2) = WALLS_BEG + wallBorder - 2;
}

/*****************************************************/
/*													 */
/*	Insert Linear Wall								 */
/*												     */
/*****************************************************/
void InsertLinearWall(u8 x, u8 y, u8 size, BOOL dir)
{
	u8* map = GetLevelPtr(x, y);
	
	for (u8 i = 0; i < size; i++)
	{
		u8 wall = ((i == 0) ? (WALLS_BEG + 1) : ((i == size - 1) ? (WALLS_BEG + 3) : (WALLS_BEG + 5)));
		if (dir == VERT)
			wall++;

		if (dir == HORIZ)
		{
			if (x + i == 0)
				wall = WALLS_BEG + 14;
			else
			if (x + i == 31)
				wall = WALLS_BEG + 12;
		}
		else
		{
			if (y + i == 0)
				wall = WALLS_BEG + 11;
			else
			if (y + i == 31)
				wall = WALLS_BEG + 13;
		}

		*(map++) = wall;
		if (dir == VERT)
			map += (MAP_CX - 1);
	}
}

/*****************************************************/
/*													 */
/*	Populate level									 */
/*												     */
/*****************************************************/
void PopulateLevel(u8 level)
{
	u8 startLevel = level / 5;
	if (startLevel > 4)
		startLevel = 4;
	
	AddLevelItems(20, MONSTER_1 + startLevel, 3);
	AddLevelItems(5, OBJ_TREASURE_GOLD, 2);
	AddLevelItems(5, OBJ_SKULL, 2);
}

/*****************************************************/
/*													 */
/*	Generate random walls							 */
/*												     */
/*****************************************************/
void GenRandomWalls(BOOL dir)
{
	for (u8 y = 3; y < 29; y += 2)
	{
		u8 x = GetRandRange(0, 6);
		u8 size = GetRandRange(2, 6);
		InsertLinearWall(dir ? x : y, dir ? y : x, size, dir);
		
		x += (size + GetRandRange(2, 9));

		if (x < 30)
		{
			size = 2 + GetRandModulo(30 - x);
			InsertLinearWall(dir ? x : y, dir ? y : x, size, dir);
		}
	}
}

/*****************************************************/
/*													 */
/*	Set exit										 */
/*												     */
/*****************************************************/
void SetExit(u8 pos)
{
	u8* map = GetLevelPtr(pos, 29);
	
	*map = WALLS_BEG;
	*(map + 1) = DGN_DOOR_H_BLUE;
	*(map + 2) = WALLS_BEG;
	*(map + MAP_CX) = DGN_DOOR_V_BLUE;	
	*(map + MAP_CX + 1) = DGN_EXIT;
	*(map + MAP_CX + 2) = DGN_DOOR_V_BLUE;
}

/*****************************************************/
/*													 */
/*	Random silver key								 */
/*												     */
/*****************************************************/
void RandPosSilverKey(u8 y)
{
	*GetLevelPtr(GetRandRange(1, 30), y) = OBJ_KEY_SILVER;
}

/*****************************************************/
/*													 */
/*	Create level 1_1								 */
/*												     */
/*****************************************************/
void CreateLevel_1_1()
{
	RandPosSilverKey(GetRandModulo(14)*2 + 2);
	SetExit(2);

	GenRandomWalls(HORIZ);
}

/*****************************************************/
/*													 */
/*	Create level 2_1								 */
/*												     */
/*****************************************************/
void CreateLevel_2_1()
{
	*GetLevelPtr(14, 5) = OBJ_KEY_SILVER;
	SetExit(28);

	GenRandomWalls(VERT);
}

/*****************************************************/
/*													 */
/*	Create level 3_1								 */
/*												     */
/*****************************************************/
void CreateLevel_3_1()
{
	SetExit(11);
	RandPosSilverKey(5);
	AddLevelItems(30, OBJ_ROCK, 2);
	AddLevelItems(30, DGN_FIRE_PATH, 1);
}

/*****************************************************/
/*													 */
/*	Create level 4_1								 */
/*												     */
/*****************************************************/
void CreateLevel_4_1()
{
	SetExit(23);
	*GetLevelPtr(5, 27) = OBJ_KEY_SILVER;

	InsertLinearWall(15, 2, 30, VERT);
	InsertLinearWall(2, 15, 28, HORIZ);
	*GetLevelPtr(15, 15) = WALLS_BEG + 15;
}

/*****************************************************/
/*													 */
/*	Create level 5_1								 */
/*												     */
/*****************************************************/
void CreateLevel_5_1()
{
	SetExit(20);
	RandPosSilverKey(6);

	for (u8 y  = 2; y < 30; y += 5)
	{
		for (u8 x = 2; x < 30; x += 5)
		{
			InsertLinearWall(x, y, GetRandRange(2, 3), (y == 27) ? HORIZ : GetRandModulo(2));
		}
	}
}

/*****************************************************/
/*													 */
/*	Create level 6_1								 */
/*												     */
/*****************************************************/
void CreateLevel_6_1()
{
	SetExit(10);
	RandPosSilverKey(10);

	AddLevelItems(60, OBJ_ROCK, 2);
	AddLevelItems(20, DGN_FIRE_PATH, 1);
}

/*****************************************************/
/*													 */
/*	Create level 7_1								 */
/*												     */
/*****************************************************/
void CreateLevel_7_1()
{
	for (u8 y = 3; y < 30; y += 4)
	{
		*GetLevelPtr(GetRandRange(1, 29), y - 1) = OBJ_KEY_SILVER;
		InsertLinearWall(0, y, 32, HORIZ);
		InsertDoorInWall(GetRandRange(1, 28), y, DGN_DOOR_H_BLUE);
	}
	
	*GetLevelPtr(GetRandRange(3, 20), 28) = OBJ_KEY_SILVER;
	SetExit(2);
}

/*****************************************************/
/*													 */
/*	Create level 8_1								 */
/*												     */
/*****************************************************/
void CreateLevel_8_1()
{
	for (u8 x = 3; x < 30; x += 4)
	{
		*GetLevelPtr(x - 1, GetRandRange(1, 29)) = OBJ_KEY_SILVER;
		InsertLinearWall(x, 0, 32, VERT);
		InsertDoorInWall(x, GetRandRange(1, 28), DGN_DOOR_V_BLUE);
	}
	
	*GetLevelPtr(28, GetRandRange(1, 20)) = OBJ_KEY_SILVER;
	SetExit(28);
}

/*****************************************************/
/*													 */
/*	Create level end game							 */
/*												     */
/*****************************************************/
void CreateLevelEndGame()
{
	u8* tile = GetLevelPtr(16, 2);
	*tile = DGN_TELEPORT_DOOR_LEFT;
	*(tile + 1) = DGN_TELEPORT_DOOR_RIGHT;

	for (u8 y = 0; y < 3; y++)
	{
		for (u8 x = 11 - y; x < 23 + y; x++)
			*GetLevelPtr(x, y + 7) = OBJ_CHEST_CLOSED + GetRandModulo(5);
	}
	
	gDungeon.startx = 16;
	gDungeon.starty = 5;
	gDungeon.palette = 2;

	_populate = FALSE;
}

/*****************************************************/
/*													 */
/*	Random generate level							 */
/*												     */
/*****************************************************/
void GenerateLevel(u8 genLevel)
{
	static const TCreateLevel _createLevels[] =	{	CreateLevel_1_1,
													CreateLevel_2_1,
													CreateLevel_3_1,
													CreateLevel_4_1,
													CreateLevel_5_1,
													CreateLevel_6_1,
													CreateLevel_7_1,
													CreateLevel_8_1,
													CreateLevelEndGame
												};
								
	TCreateLevel createLevel = _createLevels[genLevel];											
	
	cpct_memset(gLevel, 0, sizeof(gLevel));
	
	gDungeon.palette = genLevel % NB_PALETTES_WALLS;
	gDungeon.tileset = TILESET_STONE;
	gDungeon.starty = 1;
	gDungeon.startx = 1;
	_populate = TRUE;

	createLevel();

	FillBorderLevel();

	if (_populate)
		PopulateLevel(gDungeon.level);
}