//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

/** CPCTelera includes */
#include <cpctelera.h>
#include <stdio.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Game includes */
#include "..\view\view.h"
#include "..\player\player.h"
#include "..\monsters\monsters.h"
#include "..\missiles\missiles.h"

/** Tools includes */
#include "..\video\video.h"
#include "..\tools\tools.h"
#include "..\sound\sound.h"
#include "..\input\input.h"
#include "..\asm\ZX7Decrunch.h"
#include "..\asm\spriteColorize.h"
#include "..\asm\drawMapTile.h"

/** Sprites */
#include "..\sprites\impact.h"

/** Level */
#include ".\level.h"
#include ".\levels\all_levels.h"

/** Defines */
#define FIRE_RANGE			3
#define NB_COLORS_WALLS		5
#define END_LEVEL			25

/** External variable */
extern u8 gTimer;

/** Local variables */
static u8* const _levels[] = 
{ 
	G_level_1,	// 0
	NULL, 		// 1 - Level_1_1
	G_level_2,	// 2
	G_level_3,	// 3
	NULL, 		// 4 - Level_2_1
	G_level_4,	// 5	
	G_level_5,	// 6
	NULL, 		// 7 - Level_3_1
	G_level_6,	// 8
	G_level_7,	// 9 - Shop interior
	NULL, 		// 10 - Level_4_1
	G_level_8,	// 11
	G_level_9,	// 12
	NULL, 		// 13 - Level_5_1
	G_level_10,	// 14 - Shop Wood
	G_level_11,	// 15
	NULL, 		// 16 - Level_6_1
	G_level_12,	// 17
	G_level_7,	// 18 - Shop interior
	G_level_13,	// 19
	NULL, 		// 20 - Level_7_1
	G_level_14, // 21 
	G_level_15, // 22
	NULL, 		// 23 - Level_8_1
	G_level_16, // 24
	NULL		// 25 - LevelEndGame
};

static u8* const _levelTileset[] = 
{ 
	G_tiles_level_stone,
	G_tiles_level_forest
};

/** Levels color */
static const u8 _colorBackgroundLevel[] = 	{	0x56, // 0 - Sewers  - HW_GREEN
												0x46, // 1 - Dungeon - HW_CYAN
												0x57, // 2 - Dungeon - HW_SKY_BLUE
												0x58, // 3 - Dungeon - HW_MAGENTA
												0x44, // 4 - Dungeon/Forest - HW_BLUE     
												0x54, // 5 - Dungeon - HW_BLACK 
												0x5E  // 6 - Forest - HW_YELLOW
											}; 

															
static const u8 _colorWallsLevel[NB_PALETTES_WALLS][NB_COLORS_WALLS] =	
														// Pink (11) - Bright Green (12) - Cyan (13) - Bright Yellow (14) - Bright White (15)
														{	{ COLOR_RED, COLOR_BLACK, COLOR_RED, COLOR_ORANGE, COLOR_BRIGHT_RED },				// 0 - Sewers														
															{ COLOR_RED, COLOR_BLACK, COLOR_BRIGHT_RED, COLOR_BLACK, COLOR_WHITE },  			// 1 - Red
															{ COLOR_RED, COLOR_BRIGHT_RED, COLOR_BRIGHT_YELLOW, COLOR_RED, COLOR_ORANGE },		// 2 - Orange
															{ COLOR_BLUE, COLOR_BRIGHT_BLUE, COLOR_WHITE, COLOR_BLACK, COLOR_BLUE },			// 3 - Blue
															{ COLOR_RED, COLOR_BLACK, COLOR_PURPLE, COLOR_RED, COLOR_BRIGHT_RED },				// 4 - Pink
															{ COLOR_BRIGHT_CYAN, COLOR_BLUE, COLOR_BLUE, COLOR_BRIGHT_WHITE, COLOR_BRIGHT_BLUE}	// 5 - Cyan													
														};
											
/** Global level map */
u8 gLevel[MAP_SIZE];

/** Dungeon */
SDungeon gDungeon;

/** Global level tileset */
const u8 gLevelTileset[TILESET_SIZE];

/** Extern variables */
extern const u8* gTilemap;

/** Local functions */
void SetExplosion(u8* tile);
void GenerateLevel(u8 level);

/** Extern functions */
extern void InitGame();
extern void PopulateLevel(u8 level);

/*****************************************************/
/*													 */
/*	Set level color									 */
/*												     */
/*****************************************************/
void SetLevelColor(BOOL enterLevel)
{
	#define FIRST_INDEX_COLOR	15
	
	if (enterLevel && gDungeon.tileset == TILESET_STONE)
	{
		const u8* colorsWalls = (u8*)_colorWallsLevel[gDungeon.palette];

		for (u8 i = 0; i < NB_COLORS_WALLS; i++)
		{
			u8 colorWall = colorsWalls[NB_COLORS_WALLS - i - 1];
			
			asm_spriteColorize(	gLevelTileset, gLevelTileset,
								TILESET_LEVEL_CX, (u8)TILESET_LEVEL_CY, 
								FIRST_INDEX_COLOR - i, colorWall);
		}
	}
	
	cpct_setPALColour(0, _colorBackgroundLevel[gDungeon.palette]);
}

/*****************************************************/
/*													 */
/*	Get level pointer at coordinates				 */
/*												     */
/*****************************************************/
u8* GetLevelPtr(u8 x, u8 y)
{
	return gLevel + y*MAP_CX + x;
}

/*****************************************************/
/*													 */
/*	Fill border level								 */
/*												     */
/*****************************************************/
void FillBorderLevel()
{
	#define SET_LEVEL_TILE(x, tile) if (gLevel[x] == 0) gLevel[x] = tile

	u8 x;
	for (x = 1; x < MAP_CX - 1; x++)
	{
		SET_LEVEL_TILE(x, 6);
		SET_LEVEL_TILE(x + (MAP_CY - 1) * MAP_CX, 6);

		SET_LEVEL_TILE(x*MAP_CX, 7);
		SET_LEVEL_TILE(x*MAP_CX + MAP_CX - 1, 7);
	}
	
	gLevel[0] = 8;
	gLevel[MAP_CX - 1] = 9;
	gLevel[MAP_CX * MAP_CY - 1] = 10;
	gLevel[MAP_CX * (MAP_CY - 1)] = 11;
}

/*****************************************************/
/*													 */
/*	Clean level infos								 */
/*												     */
/*****************************************************/
void CleanLevelInfos()
{
	u8 i;
	for (i = 0; i < LEVEL_INFO_END; i++)
		gLevel[i] = DGN_GROUND;
}

/*****************************************************/
/*													 */
/*	Enter level										 */
/*												     */
/*****************************************************/
void EnterLevel()
{
	u8 level = gDungeon.level++;
	
	if (_levels[level] != NULL)
	{
		u8 hasBorder;
		ZX7Decrunch(_levels[level], gLevel);
	
		hasBorder = gLevel[LEVEL_BORDER];
		gDungeon.startx = gLevel[LEVEL_START_X] + 1;
		gDungeon.starty = gLevel[LEVEL_START_Y] + 1;
		gDungeon.tileset = gLevel[LEVEL_TILESET];
		gDungeon.palette = gLevel[LEVEL_PALETTE];

		if (gLevel[LEVEL_POPULATED])
			PopulateLevel(level);

		CleanLevelInfos();

		if (hasBorder)
			FillBorderLevel();
	}
	else
	{			
		GenerateLevel(gDungeon.genLevel++);
	}
		
	ZX7Decrunch(_levelTileset[gDungeon.tileset], gLevelTileset);
		
	SetLevelColor(TRUE);
	
	gPlayerInv.nbKeysSilver = 0;
	EnterPlayer();
	
	if (level == END_LEVEL)
		GameWin();	
	
	InitMonsters();
	InitMissiles();
}

/*****************************************************/
/*													 */
/*	Change level									 */
/*												     */
/*****************************************************/
void ChangeLevel()
{
	AnimationExit();
	EnterLevel();
}

/*****************************************************/
/*													 */
/*	Find element in view 							 */
/*												     */
/*****************************************************/
SCoord* FindViewElement(u8 dgnElement, u8* element)
{
	static SCoord coord;
	const u8* tileMap = gLevel;

	u8 y;
	for (y = 0; y < MAP_CY; y++)
	{
		u8 x;
		for (x = 0; x < MAP_CX; x++)
		{
			if (*tileMap == dgnElement && tileMap != element)
			{
				coord.x = x;
				coord.y = y;
				return &coord;
			}
			tileMap++;
		}
	}

	return NULL;
}

/*****************************************************/
/*													 */
/*	Open walls										 */
/*												     */
/*****************************************************/
void OpenWalls()
{
	u8* tileMap = gLevel;
	
	u16 tileIndex;
	for (tileIndex = 0; tileIndex < MAP_SIZE; tileIndex++)
	{
		u8 tileElement = *tileMap;
		if (tileElement == DGN_TRIGGER_WALL_H || tileElement == DGN_TRIGGER_WALL_V)
			*tileMap = DGN_GROUND;
		
		tileMap++;
	}
	
	PlaySFX(FX_ACTION);
}

/*****************************************************/
/*													 */
/*	Teleport 										 */
/*												     */
/*****************************************************/
BOOL Teleport(u8* srcTeleport)
{
	const SCoord* coord = FindViewElement(DGN_TELEPORTER, srcTeleport);
	if (coord != NULL)
	{
		MovePlayerAt(coord->x + 1, coord->y);
		return TRUE;
	}
	return FALSE;
}

/*****************************************************/
/*													 */
/*	Destroy Item									 */
/*												     */
/*****************************************************/
void DestroyItem(u8* tileItem)
{
	static u8 broken = 0;
	if (broken++ == ITEM_BROKEN)
	{
		u8 itemId = *tileItem;
		*tileItem = DGN_GROUND;
		
		if (itemId != OBJ_ROCK_DAMAGED)
		{
			if (itemId == OBJ_ROCK)
			{
				*tileItem = OBJ_ROCK_DAMAGED;
			}
			else
			{	
				BOOL found = (GetRandModulo(2) == 0);		
				if (found)
				{
					if (itemId == OBJ_CRATE)
					{
						u8 rand = GetRandModulo(5);
						*tileItem = (rand != 4) ? OBJ_FOOD_CORN + rand : OBJ_AMMO;
					}
					else
					{
						*tileItem = OBJ_TREASURE_GOLD;
					}
				}
			}
		}

		broken = 0;
		PlaySFX(FX_IMPACT);
	}
}

/*****************************************************/
/*													 */
/*	Get tile damages								 */
/*												     */
/*****************************************************/
u8 GetTileDamages(u8 tileId)
{
	if (tileId == DGN_FIRE_PATH)
		return FIRE_DAMAGES;
	
	if (tileId == DGN_SPIKE_TRAP && GetAnimateTile(DGN_SPIKE_TRAP) == 1)
		return SPIKE_DAMAGES;
		
	return 0;
}

/*****************************************************/
/*													 */
/*	Fill line of tile								 */
/*												     */
/*****************************************************/
void FillTileFire(u8* tile)
{
	for (u8 i = 0; i < FIRE_RANGE; i++)
	{
		u8* tileCur = tile + i;		
		u8 tileId = *tileCur;
		
		if (tileId == DGN_WATER_PATH)
			*tileCur = DGN_GROUND;
		else
		if (tileId != OBJ_HOLE_PATH)
			*tileCur = DGN_FIRE_PATH;
	}
}

/*****************************************************/
/*													 */
/*	Set explosion									 */
/*												     */
/*****************************************************/
void SetExplosion(u8* tile)
{
	static const i8 sTile[] = { -1, -MAP_CX-1, MAP_CX-1 };
	
	for (u8 i = 0; i < sizeof(sTile); i++)
		FillTileFire(tile + sTile[i]);
}

/*****************************************************/
/*													 */
/*	Set impact sprite								 */
/*												     */
/*****************************************************/
u8 GetImpactIndice(u8 status)
{
	return (status == 1) ? 1 : 0;
}

/*****************************************************/
/*													 */
/*	Search Doors to open							 */
/*												     */
/*****************************************************/
void SearchDoors(u8* doors, i8 offset)
{
	while(TRUE)
	{
		doors += offset;
		if (IS_DOOR(*doors))
			*doors = DGN_GROUND;
		else
			return;
	} 
}

/*****************************************************/
/*													 */
/*	Open Door										 */
/*												     */
/*****************************************************/
void OpenDoor(u8* door)
{
	static const i8 sDirDoors[] = { 1, -1, -MAP_CX, MAP_CX };

	for (u8 i = 0; i < sizeof(sDirDoors); i++)
		SearchDoors(door, sDirDoors[i]);

	*door = DGN_GROUND;
}

/*****************************************************/
/*													 */
/*	Game Over										 */
/*												     */
/*****************************************************/
void GameOver()
{
	WaitForFire();
	SetAllPalette(HW_BLACK);
	FlipBuffer();
	InitGame();
}

/*****************************************************/
/*													 */
/*	Game Win										 */
/*												     */
/*****************************************************/
void GameWin()
{
	u8* pvmem = GetBackScreenPtr((SCREEN_CX - TILE_CX)/2, 5 * TILE_CY);
	u8* heroWin = DeflateSpriteTile(OBJ_INVISIBLE_WALL);	
	
	EnterPlayer();
	DrawView();
	asm_drawFullMapTile(heroWin, pvmem);
	FlipBuffer();

	GameOver();
}