//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#define MISSILES_C

/** CPCTelera includes */
#include <cpctelera.h>
#include <stdio.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Game includes */
#include "..\player\player.h"
#include "..\view\view.h"
#include "..\monsters\monsters.h"
#include "..\level\level.h"
#include "..\bonus\bonus.h"

/** Tools include */
#include "..\video\video.h"
#include "..\sound\sound.h"
#include "..\tools\tools.h"

/** Sprites includes */
#include "..\sprites\dagger.h"
#include "..\sprites\fireball.h"
#include "..\sprites\impact.h"

/** Missiles include */
#include "missiles.h"

#undef MISSILES_C

/** Local defines */
#define MAX_MISSILES	6
#define MISSILE_CX		4
#define MISSILE_CY		10
#define MISSILE_SIZE	(MISSILE_CX*MISSILE_CY)

/** Local const */
static const u8 _missileDamages[] = 
{
	DAGGER_DAMAGES,
	FIREBALL_DAMAGES
};

/** Local data */
static SMissile _missiles[MAX_MISSILES];
static u8 _nbMissiles;
static SScreen _screenMissile;

/*****************************************************/
/*													 */
/*	Init Missiles									 */
/*												     */
/*****************************************************/
void InitMissiles()
{
	_nbMissiles = 0;
}

/*****************************************************/
/*													 */
/*	Convert missile position to screen				 */
/*												     */
/*****************************************************/
BOOL ConvertToScreenMissile(SPos* pos)
{
	return ConvertToScreen(pos, &_screenMissile);
}

/*****************************************************/
/*													 */
/*	Add Missile										 */
/*												     */
/*****************************************************/
void AddMissile(u8 type, SPos* pos, u8 senderId)
{
	if (_nbMissiles < MAX_MISSILES)
	{
		SMissile* newMissile = &_missiles[_nbMissiles++];
		cpct_memcpy(&newMissile->pos, pos, sizeof(SPos));
		
		newMissile->senderId = senderId; 
		newMissile->type = type;
		newMissile->infos = (type == FIREBALL || !IsPlayerBonus(PERM_THROW_MASTER)) ? FindSpriteIndex(pos->dir) : MOVE_0;
	}
}

/*****************************************************/
/*													 */
/*	Delete Missile									 */
/*												     */
/*****************************************************/
void DelMissile(SMissile* missileToDel)
{
	SMissile* lastMissile = &_missiles[--_nbMissiles];

	if (missileToDel != lastMissile)
		cpct_memcpy(missileToDel, lastMissile, sizeof(SMissile));
}

/*****************************************************/
/*													 */
/*	Damage Monster									 */
/*												     */
/*****************************************************/
void DamageMonster(SMonster* monster, SMissile* missile)
{
	u8 damages = _missileDamages[missile->type];
	
	BOOL isPlayer = (missile->senderId == PLAYER_ID);
	if (isPlayer)
		damages = IsPlayerBonus(PERM_THROW_MASTER) ? damages * BONUS_MASTER_THROW_DMG : damages;
	
	MonsterIsHit(monster, damages, isPlayer);
}

/*****************************************************/
/*													 */
/*	Missile damage Item								 */
/*												     */
/*****************************************************/
void MissileDamageItem(u8* item)
{
	u8 itemId = *item;
	if (itemId == OBJ_BOMB)
	{
		PlaySFX(FX_EXPLOSION);
		SetAllPalette(HW_BRIGHT_WHITE);
		SetExplosion(item);				
		IsPlayerCollide();
		
		Wait();
		SetGamePalette();
		SetLevelColor(FALSE);
	}
	else
	if (IsItemBreakable(itemId))
	{
		DestroyItem(item);
	}
}

/*****************************************************/
/*													 */
/*	Test if missile impact							 */
/*												     */
/*****************************************************/
void MissileImpact(SMissile* missile)
{
	u8 x, y;
	u8 missileType = missile->type;
	SPos* pos = &missile->pos;
	
	ConvertToScreenMissile(pos);
	
	x = _screenMissile.x;
	y = _screenMissile.y;
	
	/** Is player Hit */
	if (missile->senderId != PLAYER_ID && AreItemsCollide(gPlayerX, gPlayerY, x, y))
	{
		u8 damages = _missileDamages[missileType];
		PlayerIsHit(damages, FALSE);
	}
	else
	{
		/** Is Item Hit */
		u8* tileItem = GetCollidedItem(missileType, GetItemCoveredTiles(pos));
		if (tileItem == NULL || (*tileItem == OBJ_HOLE_PATH || IsTilePath(*tileItem)))
		{
			/** Is Monster Hit */
			SMonster* monster = GetCollidedMonster(missile->senderId, x, y);
			if (monster == NULL)
				return;
			
			DamageMonster(monster, missile);			
		}		
		else
			MissileDamageItem(tileItem);			
	}
	
	missile->infos = IMPACT_0;
	PlaySFX(FX_IMPACT);
}

/*****************************************************/
/*													 */
/*	Move Missile									 */
/*												     */
/*****************************************************/
void MoveMissile(SMissile* missile)
{
	SPos* pos = &missile->pos;
	u8 dir = pos->dir;

	for (u8 i = 0; i < SPEED_MISSILE; i++)
	{
		if (dir & (LEFT | RIGHT))
			MoveX(pos, dir & LEFT);

		if (dir & (TOP | DOWN))
			MoveY(pos, dir & DOWN);
			
		MissileImpact(missile);
	}
}

/*****************************************************/
/*													 */
/*	Get Missile sprite								 */
/*												     */
/*****************************************************/
u8* GetMissileSprite(SMissile* missile)
{
	static const u8 sTileMissile[TILE_SIZE];
	const u8* missileSprite;
	
	u8 status = missile->infos;
	if (status >= IMPACT_0)
	{
		u8 index = GetImpactIndice(status - IMPACT_0);
		missileSprite = gImpact_tileset[index];
	}
	else
		missileSprite = ((missile->type == DAGGER) ? g_dagger : g_fireball) + status*MISSILE_SIZE;
	
	/** Put missile Sprite into Tile */
	cpct_memcpy(sTileMissile + 3*TILE_CX, missileSprite, MISSILE_SIZE);
	
	return sTileMissile;
}

/*****************************************************/
/*													 */
/*	Draw Missile									 */
/*												     */
/*****************************************************/
void DrawMissile(SMissile* missile)
{
	#define OFFSET_DRAW_IMPACT	6

	if (ConvertToScreenMissile(&missile->pos))
	{
		u8* missileSprite = GetMissileSprite(missile);		
		
		/** Center Sprite impact */
		if (missile->infos >= IMPACT_0 && missile->pos.dir == DOWN)
		{
			if (_screenMissile.y < OFFSET_DRAW_IMPACT)
				return;
				
			_screenMissile.y -= OFFSET_DRAW_IMPACT;
		}
		
		DrawItem(&_screenMissile, missileSprite, TRUE);
	}
}

/*****************************************************/
/*													 */
/*	Update Missiles									 */
/*												     */
/*****************************************************/
void UpdateMissiles()
{
	SMissile* missile = _missiles;

	u8 i;
	for (i = 0; i < _nbMissiles; i++)
	{
		SPos* pos = &missile->pos;
		if (ConvertToScreenMissile(pos))
		{
			if (missile->infos < IMPACT_0)
			{
				if (missile->type == DAGGER && IsPlayerBonus(PERM_THROW_MASTER))
				{
					if (++missile->infos == IMPACT_0)
						missile->infos = MOVE_0;
				}
				MoveMissile(missile);
			}
			else
				missile->infos++;
		}
		else
			missile->infos = DELETE;

		if (missile->infos == DELETE)
			DelMissile(missile);

		missile++;
	}
	
	/** Draw Missiles */
	for (i = 0; i < _nbMissiles; i++)
		DrawMissile(&_missiles[i]);
}
