//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

/** CPCTelera includes */
#include <cpctelera.h>
#include <stdio.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Game includes */
#include "..\player\player.h"

/** Game includes */
#include "..\view\view.h"
#include "..\level\level.h"
#include "..\tools\tools.h"
#include "..\missiles\missiles.h"
#include "..\bonus\bonus.h"

/** Monsters includes */
#include "monsters.h"

/** Local constants */
static const u8 _damages[] = 
{
	MONST_RAT_DMG,
	MONST_GOB_DMG,
	MONST_MUMMY_DMG,
	MONST_WARLOCK_DMG,
	MONST_IMP_DMG,
	MONST_WORM_DMG,
	MONST_GREEN_CYCLOP_DMG
};

/** Locale define */
#define SET_INACTIVE(X)		X->wait = 0


/*****************************************************/
/*													 */
/*	Monster update step sprite						 */
/*												     */
/*****************************************************/
void StepMove(SMonster* monster)
{
	if (++monster->infos == MONST_DEAD_0)
		monster->infos = MONST_MOVE_0;
}

/*****************************************************/
/*													 */
/*	Monster attack player							 */
/*												     */
/*****************************************************/
void MonsterAttack(SMonster* monster)
{
	if (++monster->fightTempo > ROUND_WAIT)
	{						
		u8 monsterType = monster->type;
		BOOL attack = ((monsterType + GetRandModulo(ATTACK_LVL)) > GetPlayerAttack());
		
		if (attack)
		{
			u8 damages = _damages[monsterType];
			PlayerIsHit(damages, FALSE);
			
			/** Special attack */
			if ((monsterType == MONST_GREEN_CYCLOP || monsterType == MONST_WORM) && (GetRandModulo(3) == 0))
				AddBonus(MALUS_POISON);
		}

		StepMove(monster);		
		monster->fightTempo = 0;
	}
}

/*****************************************************/
/*													 */
/*	Test if monster collides something				 */
/*												     */
/*****************************************************/
BOOL IsMonsterCollide(SMonster* monster)
{
	u8 x, y;
	SMonster* collideMonster;
	ConvertToScreenMonster(monster);
	
	/** Collisions Tests */
	x = monster->screen.x;
	y = monster->screen.y;
	
	/** Is Monster collides Player */
	if (AreItemsCollide(gPlayerX, gPlayerY, x, y) )
	{
		MonsterAttack(monster);
		return TRUE;
	}
	
	/** Is Monster collides Item */
	{
		SPos* pos = &monster->pos;
		u8* collidedItem = GetCollidedItem(monster->type, GetItemCoveredTiles(pos));
		if (collidedItem != NULL)
		{		
			if (IsTilePath(*collidedItem))
			{
				pos->tileOn = *collidedItem;
				SET_INACTIVE(monster);
				return FALSE;
			}		
			return TRUE;
		}
		else
			pos->tileOn = DGN_GROUND;
	}
	
	/** Is Monster collides another Monster */
	collideMonster = GetCollidedMonster(monster->id, x, y);
	if (collideMonster != NULL)
	{
		SET_INACTIVE(monster);
		return TRUE;
	}
	return FALSE;
}

/*****************************************************/
/*													 */
/*	Move Monster									 */
/*												     */
/*****************************************************/
BOOL MoveMonsterDir(SMonster* monster, BOOL isX, BOOL moveDir)
{
	static SPos _monsterPos;
	SPos* pos = &monster->pos;
	
	cpct_memcpy(&_monsterPos, pos, sizeof(SPos));	
	
	if (isX)
		MoveX(pos, moveDir);
	else
		MoveY(pos, moveDir);
		
	if (IsMonsterCollide(monster))
	{
		cpct_memcpy(pos, &_monsterPos, sizeof(SPos));	
		return FALSE;
	}
	return TRUE;
}

/*****************************************************/
/*													 */
/*	Monster Fire 									 */
/*												     */
/*****************************************************/
void MonsterFire(SMonster* monster, i8 movex, i8 movey)
{
	u8 dir = NO_MOVE;
	
	u8 absMovex = GetAbsValue(movex);
	u8 absMovey = GetAbsValue(movey);
	u8 mulAbsMovex = absMovex * 2;
	
	if (absMovey > absMovex)
	{
		if (absMovey != 0)
			dir = (movey < 0) ? DOWN : TOP;

		if (absMovex != 0 && (absMovey / mulAbsMovex <= 4))
			dir |= (movex < 0 ? LEFT : RIGHT);
	}
	else
	{
		if (absMovex != 0)
			dir = (movex < 0 ? LEFT : RIGHT);

		if (absMovey != 0 && (mulAbsMovex / absMovey <= 4))
			dir |= (movey < 0) ? DOWN : TOP;
	}

	monster->pos.dir = dir;

	AddMissile(FIREBALL, &monster->pos, monster->id);
}

/*****************************************************/
/*													 */
/*	Monster Move 									 */
/*												     */
/*****************************************************/
void MoveMonster(void* monsterMoving)
{
	if (!IsPlayerBonus(BONUS_INVISIBILITY))
	{
		SMonster* monster = (SMonster*)monsterMoving;
		
		BOOL move = FALSE;
		i8 movex = gPrevPlayerX - monster->screen.x;
		i8 movey = gPrevPlayerY - monster->screen.y;
		
		if (movex != 0)
			move |= MoveMonsterDir(monster, TRUE, movex < 0);
		
		if (movey != 0)
			move |= MoveMonsterDir(monster, FALSE, movey < 0);

		if (monster->type == MONST_WARLOCK)
		{
			if (++monster->fightTempo > FIREBALL_WAIT)
			{
				MonsterFire(monster, movex, movey);
				monster->fightTempo = 0;
			}
		}
		
		if (move)
			StepMove(monster);
	}
}
