//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

/** CPCTelera includes */
#include <cpctelera.h>
#include <stdio.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Sprites includes */
#include "..\sprites\monster_1.h"
#include "..\sprites\monster_2.h"
#include "..\sprites\monster_3.h"
#include "..\sprites\monster_4.h"
#include "..\sprites\monster_5.h"
#include "..\sprites\monster_6.h"
#include "..\sprites\monster_7.h"
#include "..\sprites\hit.h"

/** Game includes */
#include "..\view\view.h"
#include "..\level\level.h"
#include "..\player\player.h"
#include "..\tools\tools.h"
#include "..\bonus\bonus.h"
#include "..\sound\sound.h"

/** Tools includes */
#include "..\asm\ZX7Decrunch.h"
#include "..\asm\flipHSprite.h"

/** Monsters includes */
#include "monsters.h"
#include "monster_move.h"

/** Local defines */
#define MAX_MONSTER			10
#define MAX_MOVING			10
#define MONSTER_DEATH 		2
#define MONSTER_XP_LVL 		20

static u8* const _compressedSprites[] = 
{
	G_monster_1, 
	G_monster_2,
	G_monster_3,
	G_monster_4,
	G_monster_5,
	G_monster_6,
	G_monster_7
};

static const u8 _xp[] =
 {
	MONST_RAT_XP,
	MONST_GOB_XP,
	MONST_MUMMY_XP,
	MONST_WARLOCK_XP,
	MONST_IMP_XP,
	MONST_WORM_XP,
	MONST_GREEN_CYCLOP_XP
};

static const u8 _life[] =
 {
	MONST_RAT_LIFE,
	MONST_GOB_LIFE,
	MONST_MUMMY_LIFE,
	MONST_WARLOCK_LIFE,
	MONST_IMP_LIFE,
	MONST_WORM_LIFE,		
	MONST_GREEN_CYCLOP_LIFE
};

/** Local data */
static SMonster _monsters[MAX_MONSTER];
static u8 _nbMonsters;
static u8 _monsterId;

/** Local Structure */
typedef struct
{
	u8 type;
	u8 sprite[TILE_SIZE * 4]; // sprites + mirrored
} SSpriteCache;

/*****************************************************/
/*													 */
/*	Init Monsters									 */
/*												     */
/*****************************************************/
void InitMonsters()
{
	_nbMonsters = 0;
	_monsterId = 0;
}

/*****************************************************/
/*													 */
/*	Convert monster position to screen				 */
/*												     */
/*****************************************************/
BOOL ConvertToScreenMonster(SMonster* monster)
{
	return ConvertToScreen(&monster->pos, &monster->screen);
}

/*****************************************************/
/*													 */
/*	Get monster sprite								 */
/*												     */
/*****************************************************/
u8* DeflateSpriteMonster(u8 monsterType)
{
	#define NB_MONSTER_BUFFER		2
	#define MONSTER_SPRITE_CY		TILE_CY*2
	#define MONSTER_SPRITE_SIZE		TILE_SIZE*2

	static SSpriteCache sSpritesMonsters[NB_MONSTER_BUFFER];
	static u8 sFreeSlot = 0;

	u8* sprite;
	SSpriteCache* spriteMonster = sSpritesMonsters;
	
	u8 monsterTypeId = monsterType + 1;
	if (spriteMonster->type == monsterTypeId)
	{
		return spriteMonster->sprite;
	}
	
	spriteMonster++;
	if (spriteMonster->type == monsterTypeId)
	{
		return spriteMonster->sprite;
	}
	
	spriteMonster = &sSpritesMonsters[sFreeSlot];		
	spriteMonster->type = monsterTypeId;
	sprite = spriteMonster->sprite;
	
	ZX7Decrunch(_compressedSprites[monsterType], sprite);
	cpct_memcpy(sprite + MONSTER_SPRITE_SIZE, sprite, MONSTER_SPRITE_SIZE);
	cpct_hflipSpriteM0(TILE_CX, MONSTER_SPRITE_CY, sprite + MONSTER_SPRITE_SIZE);

	sFreeSlot = (sFreeSlot == 0) ? 1 : 0;
	return sprite;
}

/*****************************************************/
/*													 */
/*	Add Monster										 */
/*												     */
/*****************************************************/
void AddMonster(u8 type, u8 mapx, u8 mapy)
{
	if (_nbMonsters < MAX_MONSTER)
	{
		SMonster* newMonster = &_monsters[_nbMonsters++];
		SPos* pos = &newMonster->pos;
		
		newMonster->id = ++_monsterId;
		newMonster->wait = MONSTER_ACTIVE;
		newMonster->life = _life[type];
		newMonster->type = type;
		newMonster->speed = SPEED_MOVE;
		newMonster->infos = MONST_MOVE_0;

		pos->mapx = mapx;
		pos->mapy = mapy;
		pos->y = pos->x = 0;
	}
}

/*****************************************************/
/*													 */
/*	Delete Monster									 */
/*												     */
/*****************************************************/
void DelMonster(SMonster* monsterToDel)
{
	SMonster* lastMonster = &_monsters[--_nbMonsters];
	
	if (monsterToDel != lastMonster)
		cpct_memcpy(monsterToDel, lastMonster, sizeof(SMonster));
}

/*****************************************************/
/*													 */
/*	Store Monster in level							 */
/*												     */
/*****************************************************/
void StoreMonster(SMonster* monster)
{
	u8* tile = gLevel + (monster->pos.mapx + monster->pos.mapy*MAP_CX);
	*tile = MONSTERS + monster->type;
	
	DelMonster(monster);
}

/*****************************************************/
/*													 */
/*	Create Monster in level							 */
/*												     */
/*****************************************************/
void CreateMonster(u8* tile, u8 x, u8 y)
{
	AddMonster(*tile - MONSTERS, x, y);
	*tile = DGN_GROUND;
}

/*****************************************************/
/*													 */
/*	Get Monster sprite								 */
/*												     */
/*****************************************************/
u8* GetMonsterSprite(SMonster* monster)
{
	u8 status = monster->infos;
	if (status < MONST_DEAD_0)
	{
		u8 indice = (monster->pos.dir*NB_MOVE_DIR + status)*TILE_SIZE;
		return DeflateSpriteMonster(monster->type) + indice;
	}
	else		
		return gHit_tileset[MONSTER_DEATH];
}

/*****************************************************/
/*													 */
/*	Draw Monster									 */
/*												     */
/*****************************************************/
void DrawMonster(SMonster* monster)
{
	if (ConvertToScreenMonster(monster))
	{
		SPos* pos = &monster->pos;	
		SScreen* screen = &monster->screen;
		BOOL transparent = IsTilePath(pos->tileOn);
		
		if (!IsPlayerBonus(BONUS_INVISIBILITY))
			pos->dir = (gPlayerX < screen->x); /* 0 RIGHT : 1 LEFT */
		
		DrawItem(screen, GetMonsterSprite(monster), transparent);

		if (monster->impactHit > 0)
		{
			u8 index = GetImpactIndice(monster->impactHit--);
			DrawItem(screen, gHit_tileset[index], TRUE);
		}
	}				
}

/*****************************************************/
/*													 */
/*	Monster is Hit									 */
/*												     */
/*****************************************************/
void MonsterIsHit(SMonster* monster, u8 damage, BOOL isPlayerAttack)
{
	if (monster->infos < MONST_DEAD_0)
	{
		u8 monsterType = monster->type;
		
		BOOL attack = (GetPlayerAttack() > (monsterType + GetRandModulo(ATTACK_LVL)));
		if (attack)
		{
			if (monster->life <= damage)
			{		
				if (isPlayerAttack)
				{
					u8 xp = _xp[monsterType];
					GainXP(xp);
				}
				
				monster->infos = MONST_DEAD_0;	
				PlaySFX(FX_IMPACT);
			}
			else
			{
				monster->life -= damage;

				if (monster->impactHit == 0)
					monster->impactHit = HIT_IMPACT;			
			}
		}
	}
}

/*****************************************************/
/*													 */
/*	Is Monster active								 */
/*												     */
/*****************************************************/
BOOL IsActive(SMonster* monster)
{
	if (monster->wait < MONSTER_ACTIVE)
	{
		monster->wait++;
		return FALSE;
	}
	return TRUE;
}

/*****************************************************/
/*													 */
/*	Draw Monsters									 */
/*												     */
/*****************************************************/
void DrawMonsters()
{
	for (u8 i = 0; i < _nbMonsters; i++)
		DrawMonster(&_monsters[i]);
}

/*****************************************************/
/*													 */
/*	Update Monsters									 */
/*												     */
/*****************************************************/
void UpdateMonsters()
{
	SMonster* monster = _monsters;
	for (u8 i = 0; i < _nbMonsters; i++)
	{		
		if (ConvertToScreenMonster(monster))
		{
			if (monster->infos < MONST_DEAD_0)
			{
				u8 tileOn = monster->pos.tileOn;
				u8 damages = GetTileDamages(tileOn);
			
				if (damages == 0)
					monster->speed = (tileOn == DGN_WATER_PATH) ? SPEED_MOVE_SLOW : SPEED_MOVE;
				else
					MonsterIsHit(monster, damages, TRUE);
					
				if (monster->infos != MONST_DEAD_0 && IsActive(monster))
					MoveMonster(monster);
			}
			else
			{
				if (++monster->infos == MONST_DELETE)
					DelMonster(monster);
			}
		}
		else
			StoreMonster(monster);

		monster++;
	}
	
	gPrevPlayerX = gPlayerX;
	gPrevPlayerY = gPlayerY;
}

/*****************************************************/
/*													 */
/*	Test if collision with monster					 */
/*												     */
/*****************************************************/
SMonster* GetCollidedMonster(u8 monsterId, u8 x, u8 y)
{
	SMonster* monster = _monsters;

	for (u8 i = 0; i < _nbMonsters; i++)
	{
		if (monsterId != monster->id)
		{
			ConvertToScreenMonster(monster);
			if (AreItemsCollide(x, y, monster->screen.x, monster->screen.y))
				return monster;
		}				
		monster++;
	}

	return NULL;
}

/*****************************************************/
/*													 */
/*	Burn visible monsters							 */
/*												     */
/*****************************************************/
void BurnMonsters()
{
	SMonster* monster = _monsters;	
	for (u8 i = 0; i < _nbMonsters; i++)
	{	
		if (ConvertToScreenMonster(monster))
		{
			monster->infos = MONST_DEAD_0;
			*GetLevelPtr(monster->pos.mapx, monster->pos.mapy) = DGN_FIRE_PATH;
		}
			
		monster++;
	}	
}