//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#define PLAYER_C

/** CPCTelera includes */
#include <cpctelera.h>
#include <stdio.h>

/** Structures includes */
#include "..\structures.h"
#include "..\game_const.h"

/** Game includes */
#include "..\video\video.h"
#include "..\input\input.h"
#include "..\view\view.h"
#include "..\ui\ui.h"
#include "..\ui\shop.h"
#include "..\missiles\missiles.h"
#include "..\level\level.h"
#include "..\monsters\monsters.h"
#include "..\bonus\bonus.h"
#include "..\tools\tools.h"
#include "..\sound\sound.h"

/** Sprites includes */
#include "..\sprites\hit.h"
#include "..\sprites\fx.h"
#include "..\sprites\warrior.h"

/** Tools includes */
#include "..\asm\ZX7Decrunch.h"
#include "..\asm\drawMapTile.h"

/** Player include */
#include "player.h"

#undef PLAYER_C

//#define GOD_MODE	1

/** Global Player variables */
u8 gPlayerMapx, gPlayerMapy;
u8 gPlayerX, gPlayerY;
u8 gPrevPlayerX, gPrevPlayerY;
u8 gPlayerDir;
u8 gPlayerTileOn;

SInventory gPlayerInv;
SStats gPlayerStats;
SCounters gPlayerCounters;

/** External variables */
extern SDungeon gDungeon;

/** Internal variables */
static u8 _stepAnim;

/** Internal defines */
#define KEY_GOLD		0
#define KEY_SILVER		1
#define MAX_KEYS		7

#define FX_INVISIBILITY	0
#define FX_PROTECTION	1

#define START_POS		2

/*****************************************************/
/*													 */
/*	Init player	for new game						 */
/*												     */
/*****************************************************/
void InitPlayer()
{	
	static const SStats sPlayerStatsInit = {	NO_WEAPON_DMG, NO_ARMOR_PROTECTED, 
												STATUS_OK, 
												START_LEVEL,  
												START_LIFE, START_LIFE, 
												START_FOOD, 
												START_XP, 
												START_GOLD 
											};
												
	cpct_memcpy(&gPlayerStats, &sPlayerStatsInit, sizeof(SStats));
	cpct_memset(&gPlayerInv, 0, sizeof(SInventory));
	cpct_memset(&gPlayerCounters, 0, sizeof(SCounters));
	
	CleanLife();
	DrawFullUI();
	
	gPlayerInv.selection = SEL_MISSILE;
	gDungeon.level = 0;
	gDungeon.genLevel = 0;

#ifdef GOD_MODE	
	gPlayerStats.gold = 10000;
	gPlayerStats.food = MAX_FOOD;
	gPlayerInv.nbAmmo = MAX_AMMO;
#endif
	//gPlayerStats.life = 10;
	//gPlayerInv.nbAmmo = 10;
	//AddBonus(PERM_RESURRECTION);
	//AddBonus(BONUS_INVISIBILITY);
	//AddBonus(BONUS_STRENGTH);
}

/*****************************************************/
/*													 */
/*	Player is dead									 */
/*												     */
/*****************************************************/
void PlayerDead()
{
	/** Player death */
	if (IsPlayerBonus(PERM_RESURRECTION))
	{
		RemoveBonus(PERM_RESURRECTION);

		CleanSlot(MAGIC_SLOT);
		gPlayerStats.life = gPlayerStats.maxLife;
		SetAllPalette(HW_WHITE);
		gPlayerStats.food = MAX_FOOD;
		
		EnterPlayer();
	}
	else	
	{	
		u8* pvmem = GetScreenPtr(gPlayerX, gPlayerY);
		u8* playerSprite = DeflateSpriteTile(OBJ_SKULL);
		asm_drawFullMapTile(playerSprite, pvmem);		
		cpct_setBorder(HW_BRIGHT_RED);
	
		GameOver();
	}
}

/*****************************************************/
/*													 */
/*	Enter level player								 */
/*												     */
/*****************************************************/
void EnterPlayer()
{
	SPos* pos;
	_stepAnim = 0;

	MovePlayerAt(gDungeon.startx, gDungeon.starty);
	
	pos = GetPlayerPos();
	*GetLevelPtr(pos->mapx, pos->mapy) = DGN_GROUND;
	
	gPlayerDir = TOP;
	gPlayerInv.selection = SEL_MISSILE;
	
	SetGamePalette();
	SetLevelColor(FALSE);	
	
	DrawLevelNumber();
	DrawFullUI();
}

/*****************************************************/
/*													 */
/*	Decrement value									 */
/*												     */
/*****************************************************/
u8 DecrValue(u8* valueToDecr)
{
	u8 val = *valueToDecr;
	return (val > 0) ? --(*valueToDecr) : val;
}

/*****************************************************/
/*													 */
/*	Modify attribute								 */
/*												     */
/*****************************************************/
void ModifyAttrib(i16 modif, u16* attrib, u16 max)
{
	i16 newValue = *attrib + modif;

	if (newValue < 0)
		*attrib = 0;
	else 
	if (newValue > max)
		*attrib = max;
	else
		*attrib = (u16)newValue;
}

/*****************************************************/
/*													 */
/*	Level Up										 */
/*												     */
/*****************************************************/
void LevelUp()
{
	gPlayerStats.level++;
	gPlayerStats.maxLife += LIFE_LVL_UP;
	gPlayerStats.life += LIFE_LVL_UP;
	
	SetAllPalette(HW_PASTEL_YELLOW);
	UpdateLife();
	DrawPlayerLevel();
	
	PlaySFX(FX_MAGIC);
	Wait();
	SetGamePalette();
	SetLevelColor(FALSE);
}

/*****************************************************/
/*													 */
/*	XP for next level								 */
/*												     */
/*****************************************************/
u16 GetNextLevelXP()
{
	// (n*(n+1)/2) * XP_STEP
	return (gPlayerStats.level * (gPlayerStats.level + 1) / 2) * XP_STEP;
}

/*****************************************************/
/*													 */
/*	Gain Experience									 */
/*												     */
/*****************************************************/
void GainXP(u8 xp)
{
	if (gPlayerStats.level < LEVEL_MAX)
	{
		u16 xpNext = GetNextLevelXP();
		gPlayerStats.xp += xp;
		
		if (gPlayerStats.xp >= xpNext)
		{
			LevelUp();
			gPlayerStats.xp = 0;
		}		
		
		DrawXP();
	}
}

/*****************************************************/
/*													 */
/*	Fire character									 */
/*												     */
/*****************************************************/
void FireWeaponPlayer()
{	
	if (gPlayerInv.nbAmmo > 0 && gPlayerCounters.missile == 0)
	{	
		u8 missileType = gPlayerInv.equipment[MISSILE_SLOT];
		SPos* pos = GetPlayerPos();
		AddMissile(missileType, pos, PLAYER_ID);
		
		gPlayerInv.nbAmmo--;
		gPlayerCounters.missile = RELOAD_TIME;
		
		PlaySFX(FX_MISSILE);
		DrawAmmo();
	}
}

/*****************************************************/
/*													 */
/*	Player modification life						 */
/*												     */
/*****************************************************/
void PlayerModificationLife(i16 deltaLife)
{
#ifndef GOD_MODE
	ModifyAttrib(deltaLife, &gPlayerStats.life, gPlayerStats.maxLife);
	UpdateLife();
	
	if (gPlayerStats.life == 0)
		PlayerDead();
#endif
}

/*****************************************************/
/*													 */
/*	Activate object									 */
/*												     */
/*****************************************************/
void ActivateObject(u8 objectId)
{
	switch(objectId)
	{
		case OBJ_POTION_BOMB : 			BurnMonsters(); 
		break;
		case OBJ_POTION_INVISIBILITY :	AddBonus(BONUS_INVISIBILITY); 
		break;
		case OBJ_POTION_LIFE : 			PlayerModificationLife(POTION_LIFE_BONUS); 								
		break;
		default :						AddBonus(BONUS_STRENGTH); 			
	}
	
	PlaySFX(FX_MAGIC);
}

/*****************************************************/
/*													 */
/*	Activate object from inventory slot				 */
/*												     */
/*****************************************************/
void ActivateObjectFromInv(u8 slot)
{
	u8* objectInv = &gPlayerInv.inventory[slot];
	if (*objectInv != INV_NOTHING)
	{		
		ActivateObject(*objectInv);
		*objectInv = INV_NOTHING;
		CleanSlot(INVENTORY_SLOT + slot);
	}
}

/*****************************************************/
/*													 */
/*	Activate object from inventory selection		 */
/*												     */
/*****************************************************/
void ActivateObjectFromInvSel()
{
	ActivateObjectFromInv(gPlayerInv.selection);
	gPlayerInv.selection = SEL_MISSILE;
	gPlayerCounters.missile = RELOAD_TIME;
	DrawInvSelection();
}

/*****************************************************/
/*													 */
/*	Action character								 */
/*												     */
/*****************************************************/
void ActionPlayer()
{
	if (gPlayerInv.selection == SEL_MISSILE)
		FireWeaponPlayer();
	else
		ActivateObjectFromInvSel();	
}

/*****************************************************/
/*													 */
/*	Change inventory selection						 */
/*												     */
/*****************************************************/
void ChangeInvSelection()
{
	u8 sel = (gPlayerInv.selection == SEL_MISSILE) ? SEL_INV1 : gPlayerInv.selection + 1;
	if (sel != SEL_MISSILE)
	{
		for (u8 i = 0; i < MAX_INVENTORY; i++)
		{
			if (gPlayerInv.inventory[sel] != INV_NOTHING)
			{
				gPlayerInv.selection = sel;
				break;
			}

			sel++;
		}
	}

	DrawInvSelection();
}

/*****************************************************/
/*													 */
/*	Test player tile on								 */
/*												     */
/*****************************************************/
void PlayerTileEffect()
{
	u8 playerTileOn = gPlayerTileOn;
	if (playerTileOn != DGN_GROUND)
	{
		u8 damages = GetTileDamages(playerTileOn);
		if (damages == 0)
		{
			if (gPlayerCounters.wait == 0)
				gPlayerCounters.wait = SLOWNESS;
		}
		else 
			PlayerIsHit(damages, TRUE);
	}
}

/*****************************************************/
/*													 */
/*	Update character time							 */
/*												     */
/*****************************************************/
void UpdatePlayerTimer()
{
	SCounters* playerCounter = &gPlayerCounters;
	u8* food = &playerCounter->food;

	/** Effect time */
	if (DecrValue(&playerCounter->effect) == 0)
	{
		playerCounter->effect = EFFECT_TIME;
		UpdateBonus();

		if (IsPlayerBonus(MALUS_POISON))
			PlayerIsHit(POISON_DAMAGES, TRUE);		
	}
	
	/** Wait time */
	DecrValue(&playerCounter->wait);
		
	/** Throw weapon reload time */
	DecrValue(&playerCounter->missile);

	/** Food time */
	if (DecrValue(&playerCounter->food) == 0)
	{
		playerCounter->food = FOOD_TIME;
		if (gPlayerStats.food > 0)
		{
#ifndef GOD_MODE
			gPlayerStats.food--;
#endif
			DrawFood();
		}
		else
			PlayerIsHit(FOOD_DAMAGES, TRUE);			
	}	

	if (gPlayerStats.food < FOOD_SLOWNESS && gPlayerCounters.wait == 0)
		gPlayerCounters.wait = SLOWNESS;

	/** On Tile Effect */
	PlayerTileEffect();
}

/*****************************************************/
/*													 */
/*	Get Player Sprite								 */
/*												     */
/*****************************************************/
u8* GetSpritePlayer()
{
	static const u8 sHeroSprite[HERO_CX*HERO_CY*2];
	static u8 sCurSpriteHero = 0;

	u8 spriteHero = FindSpriteIndex(gPlayerDir);
	if (spriteHero != sCurSpriteHero)
	{
		ZX7Decrunch(gWarrior[spriteHero], sHeroSprite);
		sCurSpriteHero = spriteHero;
	}

	return sHeroSprite + _stepAnim * (HERO_CX*HERO_CY);
}

/*****************************************************/
/*													 */
/*	Draw Player										 */
/*												     */
/*****************************************************/
void DrawPlayer()
{
	u8* pvmem = GetBackScreenPtr(gPlayerX, gPlayerY);
	u8* playerSprite = GetSpritePlayer();

	if (gPlayerTileOn == DGN_GROUND)
		asm_drawFullMapTile(playerSprite, pvmem);
	else
		DrawTileTransparent(playerSprite, pvmem);

	/** Draw Invisibility */
	if (IsPlayerBonus(BONUS_INVISIBILITY))
		DrawTileTransparent(gFx_tileset[FX_INVISIBILITY], pvmem);
		
	if (gPlayerCounters.impactHit > 0)
	{
		u8 index = GetImpactIndice(gPlayerCounters.impactHit--);
		DrawTileTransparent(gHit_tileset[index], pvmem);
	}
}

/*****************************************************/
/*													 */
/*	Update Step animation							 */
/*												     */
/*****************************************************/
void StepUpdate()
{
	_stepAnim = ++_stepAnim % NB_MOVE;
}

/*****************************************************/
/*													 */
/*	Animation exit level							 */
/*												     */
/*****************************************************/
void AnimationExit()
{	
	u8 i;
	SCoord* coord = FindViewElement(DGN_EXIT, NULL);
	MovePlayerAt(coord->x - 1, coord->y);
	gPlayerDir = DOWN;
		
	for (i = 0; i < TILE_CY; i++)
	{
		DrawView();
		DrawMonsters();
		
		StepUpdate();
		DrawSpriteBothBuffers(GetSpritePlayer(), gPlayerX, gPlayerY + i - 2, TILE_CX, TILE_CY - i);				
		FlipBuffer();
	}
}

/*****************************************************/
/*													 */
/*	Update player									 */
/*												     */
/*****************************************************/
void UpdatePlayer(u8 dir)
{
	if (dir & (FIRE|SELECT))
	{
		if (dir & FIRE)
			ActionPlayer();
		else
			ChangeInvSelection();
	}	
	else 
	if ((dir & MOVE_FILTER) != NO_MOVE)
	{
		gPlayerDir = dir;		
		StepUpdate();
	}
	
	DrawPlayer();
	UpdatePlayerTimer();
}

/*****************************************************/
/*													 */
/*	Add object to equipment							 */
/*												     */
/*****************************************************/
void AddObjectToEquipment(u8 slot, u8 objectId)
{
	u8 equipedObjectId = gPlayerInv.equipment[slot];
	
	gPlayerInv.equipment[slot] = objectId;
	DrawObjectInSlot(slot, objectId);
	
	if (slot == MAGIC_SLOT)
	{	
		u8 bonus = GetBonusOfObject(equipedObjectId);
		RemoveBonus(bonus);
		
		bonus = GetBonusOfObject(objectId);
		AddBonus(bonus);		
	}
}

/*****************************************************/
/*													 */
/*	Add object to inventory							 */
/*												     */
/*****************************************************/
BOOL AddObjectToInv(u8 object)
{
	if (object <= OBJ_KEY_SILVER)
	{
		if (gPlayerInv.nbKeysGold + gPlayerInv.nbKeysSilver < MAX_KEYS)		
		{
			u8* key = (object == OBJ_KEY_GOLD) ? &gPlayerInv.nbKeysGold : &gPlayerInv.nbKeysSilver;
			(*key)++;
			
			return TRUE;
		}
	}
	else
	{
		for (u8 i = 0; i < MAX_INVENTORY; i++)
		{
			if (gPlayerInv.inventory[i] == INV_NOTHING)
			{
				gPlayerInv.inventory[i] = object;
				gPlayerInv.nbObject++;

				DrawObjectInSlot(INVENTORY_SLOT + i, object);
				return TRUE;
			}			
		}
	}
	return FALSE;
}

/*****************************************************/
/*													 */
/*	Use object from inventory						 */
/*												     */
/*****************************************************/
BOOL UseObjectFromInv(u8 object)
{
	u8 i;
	for (i = 0; i < MAX_INVENTORY; i++)
	{
		if (gPlayerInv.inventory[i] == object)
		{
			gPlayerInv.inventory[i] = INV_NOTHING;
			gPlayerInv.nbObject--;

			DrawInventory();
			return TRUE;
		}			
	}
	return FALSE;
}

/*****************************************************/
/*													 */
/*	Use key	to open 								 */
/*												     */
/*****************************************************/
BOOL UseKey(u8 keyType)
{
	u8* key = (keyType == KEY_GOLD) ? &gPlayerInv.nbKeysGold : &gPlayerInv.nbKeysSilver;

	if (*key > 0)
	{
		(*key)--;

		RemoveKey();
		PlaySFX(FX_ACTION);
		return TRUE;
	}
	return FALSE;
}

/*****************************************************/
/*													 */
/*	Consume object									 */
/*												     */
/*****************************************************/
void ConsumeObject(u8 objectId)
{
	static const u8 sBonusFood[] = 
	{ 
		FOOD_STEAK, 			// OBJ_FOOD_STEAK
		FOOD_APPLE, 			// OBJ_FOOD_APPLE
		FOOD_CIDER, 			// OBJ_FOOD_CIDER
		FOOD_CORN 				// OBJ_FOOD_CORN
	};
	
	static const u8 sBonusGold[] = 
	{
		TREASURE_CHEST, 		// OBJ_TREASURE_CHEST
		TREASURE_GOLD_CUP, 		// OBJ_TREASURE_GOLD_CUP
		TREASURE_GOLD_PILES,	// OBJ_TREASURE_GOLD_PILES 
		TREASURE_GOLD 			// OBJ_TREASURE_GOLD
	}; 
	
	if (objectId == OBJ_AMMO) // Ammo
	{
		gPlayerInv.nbAmmo = MAX_AMMO;
		DrawAmmo();
	}
	else 
	if (objectId > OBJ_FOOD_STEAK) // Gold
	{
		u8 gold = sBonusGold[OBJ_TREASURE_CHEST - objectId];
		ModifyAttrib(gold, &gPlayerStats.gold, MAX_GOLD);
		DrawGold();
	}
	else // Food
	{
		u8 food = sBonusFood[OBJ_FOOD_STEAK - objectId];
		ModifyAttrib(food, &gPlayerStats.food, MAX_FOOD);
		PlayerModificationLife(food);
		DrawFood();
		DrawLife();
	}
}

/*****************************************************/
/*													 */
/*	Player is Hit									 */
/*												     */
/*****************************************************/
void PlayerIsHit(u8 damage, BOOL ignoreArmor)
{
	i8 impact;
	u8 protection = (ignoreArmor ? 0 : gPlayerStats.protection);
	
	if (IsPlayerBonus(PERM_PROTECTION))
		protection += MAGICAL_PROTECT;
	
	impact = protection - damage;
	if (impact < 0)
	{
		PlayerModificationLife(impact);
		PlaySFX(FX_IMPACT);
	}

	if (gPlayerCounters.impactHit == 0)
		gPlayerCounters.impactHit = HIT_IMPACT;
}

/*****************************************************/
/*													 */
/*	Set Damages										 */
/*												     */
/*****************************************************/
void SetWeaponDamages(u8 weaponId)
{
	static const u8 sDamages[] =
	{
		WEAPON_DAGGER,
		WEAPON_SHORT_SWORD,
		WEAPON_SWORD,
		WEAPON_SWORD_MAGIC_BLUE,
		WEAPON_SWORD_MAGIC_RED
	};
	
	gPlayerStats.damage = sDamages[weaponId - OBJ_WEAPON_DAGGER];
}

/*****************************************************/
/*													 */
/*	Set protection									 */
/*												     */
/*****************************************************/
void SetArmorProtection(u8 armorId)
{
	static const u8 sProtection[] =
	{
		ARMOR_TUNIC,
		ARMOR_LEATHER,
		ARMOR_LEATHER_STUDDED,
		ARMOR_BRONZE,
		ARMOR_MAIL,
		ARMOR_PLATE
	};
	
	gPlayerStats.protection = sProtection[armorId - OBJ_ARMOR_TUNIC];
}

/*****************************************************/
/*													 */
/*	Player attack power								 */
/*												     */
/*****************************************************/
u8 GetPlayerAttack()
{
	return (gPlayerStats.level + GetRandModulo(ATTACK_LVL)) + (IsPlayerBonus(PERM_SWORD_MASTER) ? 1 : 0);
}

/*****************************************************/
/*													 */
/*	Player interaction								 */
/*												     */
/*****************************************************/
u8* InteractPlayer(u8* object)
{
	u8 tileId = *object;
	if (tileId >= WALLS_END && tileId < MONSTERS)
	{
		/** Open closed Chest */
		if (tileId == OBJ_CHEST_CLOSED)
		{
			if (UseKey(KEY_GOLD))
			{
				*object = OBJ_TREASURE_CHEST;
			}
		}
		else
		if (tileId >= OBJECTS && tileId <= END_OBJ_TAKABLE)
		{
			/** Take objects */
			if (tileId >= OBJ_TAKABLE && tileId != OBJ_AMMO)
			{
				/** Add to equipment */
				if (tileId >= OBJ_WEAPON_DAGGER && tileId <= OBJ_MAGIC_THROW_MASTER)
				{
					u8 slot = MAGIC_SLOT;
					
					if (tileId >= OBJ_WEAPON_DAGGER && tileId <= OBJ_WEAPON_SWORD_MAGIC_RED)
					{
						slot = WEAPON_SLOT;
						SetWeaponDamages(tileId);
					}
					else
					if (tileId >= OBJ_ARMOR_TUNIC && tileId <= OBJ_ARMOR_PLATE)
					{
						slot = ARMOR_SLOT;
						SetArmorProtection(tileId);
					}

					AddObjectToEquipment(slot, tileId);
					*object = DGN_GROUND;
				}
				else
				/** Add to inventory */
				if (AddObjectToInv(tileId))
				{
					*object = DGN_GROUND;
				}
			}		
			/** Consume objects */
			else
			{
				ConsumeObject(tileId);
				*object = DGN_GROUND;
			}				
				
			PlaySFX(FX_TREASURE);
		}
		else
		{
			/** If Dungeon element */
			if (tileId < OBJECTS)
			{
				/** Use keys */
				if (IS_DOOR(tileId))
				{
					u8 keyNeeded = (tileId <= DGN_DOOR_H_RED) ? KEY_GOLD : KEY_SILVER;
					if (UseKey(keyNeeded))
					{
						OpenDoor(object);
						PlaySFX(FX_ACTION);	
					}															
				}		
				else
				/** Open Shop */			
				if (tileId == DGN_SHOP)
				{
					EnterShop();
				}
				else
				/** Change level */
				if (tileId == DGN_EXIT)
				{
					ChangeLevel();
				}
				else
				/** Change level */
				if (tileId == DGN_TELEPORT_DOOR_LEFT || tileId == DGN_TELEPORT_DOOR_RIGHT)
				{
					EnterLevel();
					return NULL;
				}
				else
				/** Open Walls */
				if (tileId == DGN_OPEN_PRESSURE_PLATE)
				{				
					OpenWalls();
					*object = DGN_USED_PRESSURE_PLATE;
					PlaySFX(FX_ACTION);
				}
				else 
				if (tileId == DGN_USED_PRESSURE_PLATE)
				{
					return NULL;
				}
			}
			else
			{				
				/** Teleport */
				if (tileId == DGN_TELEPORTER && Teleport(object))
				{
					SMonster* monster = GetCollidedMonster(PLAYER_ID, gPlayerX, gPlayerY);
					if (monster != NULL)				
						monster->infos = MONST_DEAD_0;
					
					PlaySFX(FX_MAGIC);
					return NULL;
				}
				else 			
				/** Walk on special tile */
				if (IsTilePath(tileId))
				{
					gPlayerTileOn = tileId;
					return NULL;
				}
				else
				/** Destroy Item */
				if (IsItemBreakable(tileId))
				{
					DestroyItem(object);
				}
			}
		}
	}

	return object;
}

/*****************************************************/
/*													 */
/*	Player collision								 */
/*												     */
/*****************************************************/
BOOL IsPlayerCollide()
{
	u8* itemCollide;
	SMonster* monster = GetCollidedMonster(PLAYER_ID, gPlayerX, gPlayerY);
	if (monster != NULL)
	{
		u8 damages = gPlayerStats.damage;
#ifndef GOD_MODE			
		if (IsPlayerBonus(BONUS_STRENGTH))
			damages += MAGICAL_STRENGTH;
#else	
		damages = 255;
#endif
		MonsterIsHit(monster, damages, TRUE);
		PlaySFX(FX_IMPACT);
		return TRUE;
	}
	
	itemCollide = GetCollidedItem(PLAYER, GetPlayerCoveredTiles());
	return (itemCollide != NULL);
}

/*****************************************************/
/*													 */
/*	Check player move								 */
/*												     */
/*****************************************************/
u8 MovePlayer()
{
	u8 move = CheckUserInput();
	if (move != NO_MOVE)
	{
		if (move >= SEL1)
		{
			ActivateObjectFromInv(move - SEL1);
			move = NO_MOVE;
		}
		else
		{
			if (!(move & FIRE))
			{
				if (gPlayerCounters.wait <= 1)
					ScrollView(move);
				else
					move = NO_MOVE;
			}
		}
	}
	return move;
}
