//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

/** CPCTelera includes */
#include <cpctelera.h>

/** Structures includes */
#include "..\game_const.h"
#include "..\structures.h"

/** Game includes */
#include "..\video\video.h"
#include "..\view\view.h"
#include "..\fonts\fontsHelper.h"
#include "..\ui\ui.h"
#include "..\player\player.h"
#include "..\input\input.h"

/** Sprites includes */
#include "..\sprites\shop.h"
#include "..\sprites\arrow.h"
#include "..\sprites\exit.h"

/** Tools includes */
#include "..\tools\tools.h"
#include "..\asm\ZX7Decrunch.h"

/** Local define */
#define SHOP_CX		(56/2)
#define SHOP_CY		48

#define SHOP_X		(SCREEN_CX - SHOP_CX) / 2
#define SHOP_Y		(VIEW_HEIGHT - SHOP_CY) / 2

#define SELL_CX		(((SHOP_CX - 4) - TILE_CX*3)/ 4)
#define SELL_CY		TILE_CY
#define SELL_X		(SHOP_X + 2)
#define SELL_Y		(SHOP_Y + 6)

#define ARROW_X		(SHOP_X)
#define ARROW_Y		(SELL_Y + TILE_CY + 2)

#define NB_OBJ		3
#define NB_SEL		4
#define PREV_SEL	0
#define NEXT_SEL	4
#define EXIT_SEL	5

#define NB_DIGIT	4
#define PRICES_X	(SHOP_X + 20/2)
#define PRICES_Y	(SHOP_Y + 33)
#define PRICES_CX 	(NB_DIGIT*FONTS_CX)

#define EXIT_X		(SHOP_X + 24/2)
#define EXIT_Y		(SHOP_Y + 40)
#define EXIT_CX		G_EXIT_0_W
#define EXIT_CY		G_EXIT_0_H

#define POTIONS		6

#define GOLD_COEF	100

static u8 _objectType;
static u8 _currentSel;

enum
{
	ARMOR,
	WEAPON,
	POTION,
	MAGIC
};

static const u8 _objToSell[] 		=	{	0, 0, 0, 
											0, 0, 0, 
											OBJ_POTION_BOMB, OBJ_POTION_INVISIBILITY, OBJ_POTION_LIFE, 
											OBJ_MAGIC_PROTECTION, OBJ_MAGIC_THROW_MASTER, OBJ_MAGIC_SWORD_MASTER };

static const u8 _pricesArmor[]		=	{ 	ARMOR_TUNIC_GOLD, 
											ARMOR_LEATHER_GOLD, 
											ARMOR_LEATHER_STUDDED_GOLD, 
											ARMOR_MAIL_GOLD, 
											ARMOR_BRONZE_GOLD,
											ARMOR_PLATE_GOLD };
											
static const u8 _pricesWeapon[]	=		{ 	WEAPON_DAGGER_GOLD, 
											WEAPON_SHORT_SWORD_GOLD, 
											WEAPON_SWORD_GOLD, 
											WEAPON_SWORD_MAGIC_BLUE_GOLD, 
											WEAPON_SWORD_MAGIC_RED_GOLD };
											
static const u8 _pricesPotion[]	=		{	POTION_BOMB_GOLD, 
											POTION_INVISIBILITY_GOLD, 
											POTION_LIFE_GOLD };
											
static const u8 _pricesMagic[]		=	{	PERM_PROTECTION_GOLD,
											PERM_THROW_MASTER_GOLD,
											PERM_SWORD_MASTER_GOLD };

#define COMPUTE_POS(X)	(SELL_X + SELL_CX + ((X) * (SELL_CX + TILE_CX)))
						
/*****************************************************/
/*													 */
/*	Get Selection									 */
/*												     */
/*****************************************************/
u8 GetSelection()
{
	return _objectType * NB_OBJ + _currentSel - 1;
}

/*****************************************************/
/*													 */
/*	Draw Exit Sprite								 */
/*												     */
/*****************************************************/
void DrawExit(BOOL open)
{
	u8* screen = GetScreenPtr(EXIT_X, EXIT_Y);
	cpct_drawSprite( open ? g_exit_1 : g_exit_0, screen, EXIT_CX, EXIT_CY);
}

/*****************************************************/
/*													 */
/*	Draw Table										 */
/*												     */
/*****************************************************/
void DrawTable()
{
	u8* toScreen = GetScreenPtr(SHOP_X, SHOP_Y);
	u8* decomp = DecompToBackBuffer(G_shop);
	
	cpct_drawSprite(decomp, toScreen, SHOP_CX, SHOP_CY);
	
	DrawExit(TRUE);
}

/*****************************************************/
/*													 */
/*	Get price of selection							 */
/*												     */
/*****************************************************/
u16 GetPrice()
{
	u8 selection = GetSelection();
	u8 selElement = _objToSell[selection];
	u8 indice = selection % NB_OBJ;
	
	u8 *prices = _pricesMagic;
	if (selElement >= OBJ_ARMOR_TUNIC && selElement <= OBJ_ARMOR_PLATE)
	{
		prices = _pricesArmor;
		indice = selElement - OBJ_ARMOR_TUNIC;
	}
	else
	if (selElement >= OBJ_WEAPON_DAGGER && selElement <= OBJ_WEAPON_SWORD_MAGIC_RED)
	{
		prices = _pricesWeapon;
		indice = selElement - OBJ_WEAPON_DAGGER;
	}
	else	
	if (selElement >= OBJ_POTION_BOMB && selElement <= OBJ_POTION_STRENGTH)
	{
		prices = _pricesPotion;
	}
		
	return prices[indice] * GOLD_COEF;
}

/*****************************************************/
/*													 */
/*	Clean Price	 									 */
/*												     */
/*****************************************************/
void CleanPrice()
{
	u8* screen = GetScreenPtr(PRICES_X, PRICES_Y);
	cpct_drawSolidBox(screen, cpctm_px2byteM0(COLOR_WHITE, COLOR_WHITE), PRICES_CX, FONTS_CY);
}

/*****************************************************/
/*													 */
/*	Draw Price	 									 */
/*												     */
/*****************************************************/
void DrawPrice(u8 pos)
{
	CleanPrice();

	if (_objToSell[pos] != INV_NOTHING)
	{
		if (_currentSel > PREV_SEL && _currentSel < NEXT_SEL)
			PrintInt(PRICES_X, PRICES_Y, GetPrice(), NB_DIGIT, SCREEN_BUFFER);
	}
}

/*****************************************************/
/*													 */
/*	Draw Selection 									 */
/*												     */
/*****************************************************/
void DrawShopSelection(BOOL draw)
{
	u8* screen;
	u8 pos;
	
	if (_currentSel == PREV_SEL)
		pos = SHOP_X;
	else 
	if (_currentSel == NEXT_SEL)
		pos = SHOP_X + SHOP_CX - G_ARROW_W;
	else
		pos = COMPUTE_POS(_currentSel - 1);
		
	screen = GetScreenPtr(pos, ARROW_Y);
	
	if (draw)
	{
		cpct_drawSprite(g_arrow, screen, G_ARROW_W, G_ARROW_H);	
		DrawPrice(GetSelection());
	}
	else
		cpct_drawSolidBox(screen, cpctm_px2byteM0(COLOR_RED, COLOR_RED), G_ARROW_W, G_ARROW_H);
}

/*****************************************************/
/*													 */
/*	Clean Items										 */
/*												     */
/*****************************************************/
void CleanItem(u8* screen)
{
	cpct_drawSolidBox(screen, cpctm_px2byteM0(COLOR_SKY_BLUE, COLOR_SKY_BLUE), TILE_CX, TILE_CY);
}

/*****************************************************/
/*													 */
/*	Draw Items										 */
/*												     */
/*****************************************************/
void DrawItems()
{
	for (u8 i = 0; i < NB_OBJ; i++)
	{
		u8 item = _objToSell[_objectType * NB_OBJ + i];
		u8* screen = GetScreenPtr(COMPUTE_POS(i), SELL_Y);	
		
		CleanItem(screen);
		if (item != INV_NOTHING)
		{
			const u8* sprite = DeflateSpriteTile(item);
			DrawTileTransparent(sprite, screen);				
		}				
	}
}

/*****************************************************/
/*													 */
/*	Buy Item										 */
/*												     */
/*****************************************************/
void BuyItem()
{
	u8 selection = GetSelection();
	const u8* objectBuy = &_objToSell[selection];
	
	BOOL hasNoLeft = (_objectType == POTION && gPlayerInv.nbObject == MAX_INVENTORY);
	
	if (hasNoLeft == FALSE && *objectBuy != INV_NOTHING)
	{	
		u16 price = GetPrice();
		if (gPlayerStats.gold > price)
		{						
			u8* screen = GetScreenPtr(COMPUTE_POS(_currentSel - 1), SELL_Y);
			CleanItem(screen);
		
			gPlayerStats.gold -= price;
			DrawGold();
			InteractPlayer((u8*)objectBuy);
		}
	}
}
		
/*****************************************************/
/*													 */
/*	Select Item										 */
/*												     */
/*****************************************************/
void SelectItem()
{
	u8 curSel = _currentSel;
	
	if (curSel == NEXT_SEL)
	{			
		if (_objectType++ == MAGIC)
			_objectType = ARMOR;
			
		DrawItems();			
	}
	else if (curSel == PREV_SEL)
	{
		if (_objectType-- == ARMOR)
			_objectType = MAGIC;		
			
		DrawItems();
	}
	else
		BuyItem();
}

/*****************************************************/
/*													 */
/*	Close shop										 */
/*												     */
/*****************************************************/
void CloseShop()
{	
	DrawExit(FALSE);

	DrawShopSelection(FALSE);
	CleanPrice();
	
	_currentSel = EXIT_SEL;
}

/*****************************************************/
/*													 */
/*	Exit shop										 */
/*												     */
/*****************************************************/
void ExitShop()
{
	CopyScreenToBackBuffer();		
	DrawFullUI();
}

/*****************************************************/
/*													 */
/*	Select object to buy							 */
/*												     */
/*****************************************************/
void MakeSelection()
{
	u8 prevInput = NO_MOVE;
	DrawShopSelection(TRUE);
	
	while(TRUE)
	{
		u8 input = CheckUserInput();
		if (prevInput != input)
		{
			prevInput = input;
			if (input != NO_MOVE)
			{
				if (input == TOP)
				{
					CloseShop();
				}
				else
				{								
					if (input == FIRE)
					{
						if (_currentSel == EXIT_SEL)
						{
							ExitShop();
							break;
						}
					
						SelectItem();
					}
					
					if (_currentSel == EXIT_SEL)
					{
						DrawExit(TRUE);
						DrawItems();
						_currentSel = PREV_SEL;
					}
				
					DrawShopSelection(FALSE);
				
					if (input == RIGHT && _currentSel < NB_SEL)
						_currentSel++;
					else 
					if (input == LEFT && _currentSel > PREV_SEL)
						_currentSel--;		
												
					DrawShopSelection(TRUE);
				}
			}
		}
	}
}

/*****************************************************/
/*													 */
/* Fill object to sell								 */
/*												     */
/*****************************************************/
void FillShelve()
{
	static const u8 sPotions[] = {OBJ_POTION_BOMB, OBJ_POTION_INVISIBILITY, OBJ_POTION_LIFE};

	u8 i, eqt;
	u8* sell = _objToSell;
	
	/** Fill Armors */
	eqt = gPlayerInv.equipment[ARMOR_SLOT];
	if (eqt == INV_NOTHING)
		eqt = OBJ_ARMOR_TUNIC - 1;
	
	for (i = 1; i < NB_OBJ + 1; i++)
	{
		u8 item = eqt + i;
		
		*sell = INV_NOTHING;
		if (item <= OBJ_ARMOR_PLATE)
			*sell = item;
		
		sell++;
	}
	
	/** Fill Weapons */
	eqt = gPlayerInv.equipment[WEAPON_SLOT];
	if (eqt == INV_NOTHING)
		eqt = OBJ_WEAPON_DAGGER - 1;

	for (i = 1; i < NB_OBJ + 1; i++)
	{
		u8 item = eqt + i;
		
		*sell = INV_NOTHING;
		if (item <= OBJ_WEAPON_SWORD_MAGIC_RED)
			*sell = item;
			
		sell++;
	}
	
	/** Fill Potions */
	cpct_memcpy(_objToSell + POTIONS, sPotions, NB_OBJ);
}

/*****************************************************/
/*													 */
/*	Enter Shop										 */
/*												     */
/*****************************************************/
void EnterShop()
{
	_objectType = ARMOR;
	_currentSel = PREV_SEL;
	
	DrawTable();
	FillShelve();
	DrawItems();
	MakeSelection();
}
