//-----------------------------LICENSE NOTICE------------------------------------
//  This file is part of Deeper Warrens
//  Copyright (C) 2018 Arnaud Bouche (@Arnaud6128)
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

#define VIDEO_C

/** CPCTelera includes */
#include <cpctelera.h>

/** Structures includes */
#include "..\structures.h"

/** Video include */
#include "video.h"

/** Game includes */
#include "..\view\view.h"

/** Tools includes */
#include "..\asm\ZX7Decrunch.h"

#undef VIDEO_C

// Memory location definition
enum
{
	VIDEO_MEM,
	BUFFER_MEM,
	NB_VIDEO_BUFFER
};

/** Declare mask table */
cpctm_createTransparentMaskTable(gMaskTable, MASK_TABLE_LOC, M0, 0);

/** Internal variables */
static u8 _VMem;				// Current video memory

/*****************************************************/
/*													 */
/*	Flip double buffer								 */
/*												     */
/*****************************************************/
void FlipBuffer()
{
	cpct_waitVSYNC();

	/** Change memory page */
	if (_VMem == BUFFER_MEM)
	{
		cpct_setVideoMemoryPage(cpct_pageC0);
		_VMem = VIDEO_MEM;
	}
	else
	{
		cpct_setVideoMemoryPage(cpct_page80);
		_VMem = BUFFER_MEM;
	}
}

/*****************************************************/
/*													 */
/*	Get video buffer pointer						 */
/*												     */
/*****************************************************/
u8* GetScreenPtr(u8 xPos, u8 yPos)
{
	return cpct_getScreenPtr(_VMem == VIDEO_MEM ? (u8*)CPCT_VMEM_START : (u8*)SCREEN_BUFF_LOC, xPos, yPos);
}

/*****************************************************/
/*													 */
/*	Get back buffer pointer							 */
/*												     */
/*****************************************************/
u8* GetBackScreenPtr(u8 xPos, u8 yPos)
{
	return cpct_getScreenPtr(_VMem != VIDEO_MEM ? (u8*)CPCT_VMEM_START : (u8*)SCREEN_BUFF_LOC, xPos, yPos);
}

/*****************************************************/
/*													 */
/*	Get back buffer									 */
/*												     */
/*****************************************************/
u8* GetBackBufferPtr()
{
#ifndef WINCPCTELERA
	return (_VMem != VIDEO_MEM) ? (u8*)CPCT_VMEM_START : (u8*)SCREEN_BUFF_LOC;
#else
	return (_VMem != VIDEO_MEM) ? cpct_getScreenPtr(CPCT_VMEM_START, 0, 0) : cpct_getScreenPtr(SCREEN_BUFF_LOC, 0, 0);
#endif
}

/*****************************************************/
/*													 */
/*	Copy Screen to BackBuffer						 */
/*												     */
/*****************************************************/
void CopyScreenToBackBuffer()
{
	u8* screenBuffer = (_VMem == VIDEO_MEM) ? (u8*)CPCT_VMEM_START : (u8*)SCREEN_BUFF_LOC;
	cpct_memcpy(GetBackBufferPtr(), screenBuffer, VMEM_SIZE);
}

/*****************************************************/
/*													 */
/*	Draw tile Transparent							 */
/*												     */
/*****************************************************/
void DrawTileTransparent(const u8* sprite, const u8* vmem)
{
	cpct_drawSpriteMaskedAlignedTable(sprite, vmem, TILE_CX, TILE_CY, gMaskTable);
}

/*****************************************************/
/*													 */
/*	Draw tile on both screen buffers				 */
/*												     */
/*****************************************************/
void DrawTileTransparentBothBuffers(const u8* sprite, u8 x, u8 y)
{
	DrawTileTransparent(sprite, GetScreenPtr(x, y));
	DrawTileTransparent(sprite, GetBackScreenPtr(x, y));
}

/*****************************************************/
/*													 */
/*	Draw tile on both screen buffers				 */
/*												     */
/*****************************************************/
void DrawSpriteBothBuffers(const u8* sprite, u8 x, u8 y, u8 cx, u8 cy)
{	
	cpct_drawSpriteMaskedAlignedTable(sprite, GetScreenPtr(x, y), cx, cy, gMaskTable);
	cpct_drawSpriteMaskedAlignedTable(sprite, GetBackScreenPtr(x, y), cx, cy, gMaskTable);
}

/*****************************************************/
/*													 */
/*	Draw Solid Box on both screen buffers			 */
/*												     */
/*****************************************************/
void DrawSolidBoxBothBuffers(u8 color, u8 x, u8 y, u8 cx, u8 cy)
{
	for (u8 i = 0; i < NB_VIDEO_BUFFER; i++)
	{
		u8* vmem = (i == VIDEO_MEM) ? GetScreenPtr(x, y) : GetBackScreenPtr(x, y);
		cpct_drawSolidBox(vmem, color, cx, cy);
	}
}

/*****************************************************/
/*													 */
/*	Draw line both buffer							 */
/*												     */
/*****************************************************/
void DrawLineBothBuffer(u8 x, u8 y, u8 sizeInPix, u8 maxInPix, u8 color, u8 colorBack)
{
	for (u8 i = 0; i < NB_VIDEO_BUFFER; i++)
	{
		u8* vmem = (i == VIDEO_MEM) ? GetBackScreenPtr(x, y) : GetScreenPtr(x, y);
		DrawLine(vmem, sizeInPix, maxInPix, color, colorBack);
	}
}

/*****************************************************/
/*													 */
/*	Draw line										 */
/*												     */
/*****************************************************/
void DrawLine(u8* vmem, u8 valueInPix, u8 maxInPix, u8 color, u8 colorBack)
{
	u8 valueInByte = valueInPix / 2;
	u8 maxInByte = maxInPix / 2;

	cpct_memset(vmem, cpctm_px2byteM0(colorBack, colorBack), maxInByte);

	if (valueInByte > 0)
	{	
		u8* vmemColor = vmem;
		
		u8 i;
		for (i = 0; i < valueInByte; i++)
			*(vmemColor++) = cpctm_px2byteM0(color, color);
	}
	
	if (valueInPix % 2)
		*(vmem + valueInByte) = cpctm_px2byteM0(color, colorBack);
}

/*****************************************************/
/*													 */
/*	Set game palette								 */
/*												     */
/*****************************************************/
void SetGamePalette()
{
	static const u8 sGamePalette[] = 
	{ 
		0x5E, // 00 : HW_YELLOW			(12)
		0x54, // 01 : HW_BLACK			(0)
		0x44, // 02 : HW_BLUE			(1)
		0x55, // 03 : HW_BRIGHT_BLUE	(2)
		0x5c, // 04 : HW_RED			(3)
		0x4c, // 05 : HW_BRIGHT_RED 	(6)
		0x45, // 06 : HW_PURPLE			(7)
		0x56, // 07 : HW_GREEN			(9)
		0x57, // 08 : HW_SKY_BLUE		(11)
		0x40, // 09 : HW_WHITE			(13)
		0x4e, // 10 : HW_ORANGE			(15)
		0x47, // 11 : HW_PINK			(16)
		0x52, // 12 : HW_BRIGHT_GREEN	(18)
		0x53, // 13 : HW_BRIGHT_CYAN	(20)		
		0x4a, // 14 : HW_BRIGHT_YELLOW	(24)
		0x4b  // 15 : HW_BRIGHT_WHITE	(26)
	};

	/** Set colors */
	cpct_setBorder(HW_BLACK);
	cpct_setPalette(sGamePalette, sizeof(sGamePalette));
}

/*****************************************************/
/*													 */
/*	Set palette in same color						 */
/*												     */
/*****************************************************/
void SetAllPalette(u8 color)
{
	u8 i;
	for (i = 0; i < 17; i++)
		cpct_setPALColour(i, color);
}

/*****************************************************/
/*													 */
/*	Decompress and Draw to screen					 */
/*												     */
/*****************************************************/
u8* DecompToBackBuffer(u8* compress)
{
	u8* decomp = GetBackBufferPtr();
	ZX7Decrunch(compress, decomp);
	return decomp;
}